package org.eclipse.tm.tcf.testservice;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.tm.tcf.extender.peer.IPeer;
import org.eclipse.tm.tcf.extender.service.IServiceUtilities;
import org.eclipse.tm.tcf.extender.service.ITaskUtilities;
import org.eclipse.tm.tcf.extender.service.provider.IServiceProvider;
import org.eclipse.tm.tcf.extender.service.provider.IServiceProviderInterface;
import org.eclipse.tm.tcf.extender.service.provider.IServiceProviderInterface.ICommandServer;
import org.eclipse.tm.tcf.extender.service.provider.IServiceProviderManager;
import org.eclipse.tm.tcf.extensions.TCFTCP;

public class TestServiceProviderManager implements IServiceProviderManager {
	
	private static final int WAIT_TIME = 10;
	private static final int NB_WAIT_CYCLES = 50;
	
	private static String status = TestServiceStatus.AVAIABLE_STATE;;
	private static final Object lock = new Object();
	
	private final Map<IPeer, TestServiceProvider> providers = new HashMap<IPeer, TestServiceProvider>();
	private ITaskUtilities taskUtilities;
	
	public void createServiceProvider(IPeer peer,
			IServiceProviderInterface serviceInterface) {
		TestServiceProvider provider = new TestServiceProvider(serviceInterface);
		providers.put(peer, provider);
	}

	public String getType() {
		return TCFTCP.TCP_TYPE;
	}

	public void removeServiceProvider(IPeer peer) {
		TestServiceProvider provider = providers.remove(peer);
		provider.terminate();
	}

	public void setServiceUtilities(IServiceUtilities utilities) {
		// TODO Auto-generated method stub

	}

	public void setTaskUtilities(ITaskUtilities utilities) {
		taskUtilities = utilities;

	}
	
	private class TestServiceProvider implements IServiceProvider, ITestService, ICommandServer  {
		
		private TestInstruction[] instructions;
		private TestResult[] results;
		private int index;
		private int cnt;
		private String token;
		private final IServiceProviderInterface providerInterface; 
		
		public TestServiceProvider(IServiceProviderInterface serviceInterface){
			super();
			providerInterface = serviceInterface;
			instructions = new TestInstruction[0];
			results = new TestResult[0];
			index = 0;
			token = "";
			cnt = 0;
			providerInterface.addCommandServer(NAME, this);
		}
		
		public void terminate(){
			providerInterface.removeCommandServer(NAME, this);
		}

		public String getName() {
			return NAME;
		}

		public void command(final String token, final String name, final Object[] data) {
			synchronized (lock) {
				if (name.equals(RUN_TEST_COMMAND)) {
					if (status.equals(TestServiceStatus.AVAIABLE_STATE)) {
						changeStatus(TestServiceStatus.BUSY_STATE);
						taskUtilities.invokeLater(new Runnable() {
							public void run() {
								initTests(token, data);
							}
						});
					} else {
						taskUtilities.invokeLater(new Runnable() {
							public void run() {
								providerInterface.sendResult(token,
										new Object[] { ERROR_TAG,
												"The service is busy" });
							}
						});
					}
				} else if (name.equals(GET_STATUS_COMMAND)) {
					taskUtilities.invokeLater(new Runnable() {
						public void run() {
							TestServiceStatus s = new TestServiceStatus(status);
							providerInterface.sendEvent(NAME, STATUS_EVENT,
									new Object[] { s.toMap() });
						}
					});
				}
			}
		}
		
		private void changeStatus(final String st){
			synchronized (lock) {
				status = st;
				taskUtilities.invokeLater(new Runnable() {
					public void run() {
						TestServiceStatus s = new TestServiceStatus(status);
						providerInterface.sendEvent(NAME, STATUS_EVENT,
								new Object[] { s.toMap() });
					}
				});
			}
		}
		
		@SuppressWarnings("unchecked")
		private void initTests(final String token,final Object[] data){
			if(data.length==0){
				taskUtilities.invokeLater(new Runnable(){
					public void run() {
						providerInterface.sendResult(token, new Object[]{ERROR_TAG, "There is a no test"});
					}
				});
				index = 0;
				changeStatus(TestServiceStatus.AVAIABLE_STATE);
			} else {
				this.instructions = new TestInstruction[data.length];
				for(int i=0; i<data.length; i++){
					this.instructions[i] = new TestInstruction((Map)data[i]);
				}
				this.results = new TestResult[data.length];
				this.index = 0;
				this.token = token;
				taskUtilities.invokeLater(new Runnable(){
					public void run() {
						waitTest();
					}
				});
			}
		}
		
		private void waitTest(){
			/* Emulate long asynchrone operation */ 
			if(cnt >= NB_WAIT_CYCLES){
				taskUtilities.invokeLater(new Runnable(){
					public void run() {
						runTest();
					}
				});
			} else {
				try {
					Thread.sleep(WAIT_TIME);
					cnt++;
					taskUtilities.invokeLater(new Runnable(){
						public void run() {
							waitTest();
						}
					});
				} catch (InterruptedException e) {
					finishTestsWithError("Internal Error");
				}
			}
		}
		
		private void runTest(){
			TestInstruction instruction = instructions[index];
			if(instruction.getOperation().equals(TestInstruction.ADD_INSTRUCTION)) {results[index] = new TestResult(instruction.getId(), (instruction.getArg1() + instruction.getArg2()), false, false, "");}
			else if(instruction.getOperation().equals(TestInstruction.SUB_INSTRUCTION)) {results[index] = new TestResult(instruction.getId(), (instruction.getArg1() - instruction.getArg2()), false, false, "");}
			else if(instruction.getOperation().equals(TestInstruction.MUL_INSTRUCTION)) {results[index] = new TestResult(instruction.getId(), (instruction.getArg1() * instruction.getArg2()), false, false, "");}
			else if(instruction.getOperation().equals(TestInstruction.DIV_INSTRUCTION)) {results[index] = new TestResult(instruction.getId(), (instruction.getArg1() / instruction.getArg2()-1), false, false, "");}
			else if(instruction.getOperation().equals(TestInstruction.INT_ERROR_OPERATION)) {finishTestsWithError("Internal Error");return;}
			else {results[index] = new TestResult(instruction.getId(), 0, true, false, "Unsuported operation : "+instruction.getOperation());}
			final TestResult result = results[index];
			taskUtilities.invokeLater(new Runnable(){
				public void run() {
					providerInterface.sendProgress(token, new Object[]{result.toMap()});
				}
			});
			index++;
			cnt = 0;
			if(index == instructions.length){
				taskUtilities.invokeLater(new Runnable(){
					public void run() {
						finishTests();
					}
				});
			} else {
				taskUtilities.invokeLater(new Runnable(){
					public void run() {
						waitTest();
					}
				});
			}
		}
		
		private void finishTestsWithError(final String message){
			taskUtilities.invokeLater(new Runnable(){
				public void run() {
					providerInterface.sendResult(token, new Object[]{ERROR_TAG, message});
				}
			});
			index = 0;
			changeStatus(TestServiceStatus.AVAIABLE_STATE);
		}
		

		
		private void finishTests(){
			taskUtilities.invokeLater(new Runnable(){
				public void run() {
					providerInterface.sendResult(token, new Object[0]);
				}
			});
			index = 0;
			changeStatus(TestServiceStatus.AVAIABLE_STATE);
			
		}
		
	}

}
