package org.eclipse.tm.tcf.extensions.channel;

import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.LinkedList;

import org.eclipse.tm.tcf.extender.channel.IChannel;
import org.eclipse.tm.tcf.extender.channel.IChannelInterface;
import org.eclipse.tm.tcf.extender.channel.IChannelInterface.IMessageSender;
import org.eclipse.tm.tcf.extender.message.EndOfStream;
import org.eclipse.tm.tcf.extender.message.IMessage;
import org.eclipse.tm.tcf.extender.peer.IPeer;
import org.eclipse.tm.tcf.extender.service.ITaskUtilities;
import org.eclipse.tm.tcf.user.TCF;

public abstract class AbstractChannel implements IMessageSender, IChannel {

	protected IPeer peer;
	protected IChannelInterface channelInterface;
	protected ITaskUtilities taskUtilities;
	protected Thread inThread;
	protected Thread outThread;
	protected LinkedList<IMessage> outMessages = new LinkedList<IMessage>();

	public AbstractChannel(IPeer peer, IChannelInterface channelInterface,
			ITaskUtilities taskUtilities) {
		this.peer = peer;
		this.channelInterface = channelInterface;
		channelInterface.addMessageSender(this);
		this.taskUtilities = taskUtilities;
		inThread = new Thread() {
			public void run() {
				inProcess();
			}
		};
		inThread.setName("TCF Input Process of peer " + peer.getName());
		outThread = new Thread() {
			public void run() {
				outProcess();
			}
		};
		outThread.setName("TCF Output Process of peer " + peer.getName());

	}

	public void start() throws Exception {
		inThread.start();
		outThread.start();
	}

	public void stop() throws Exception {
		channelInterface.removeMessageSender(this);
		sendMessage(new EndOfStream());
		inThread.join();
		outThread.join();
	}

	private void inProcess() {
		assert Thread.currentThread() == inThread;
		try {
			initRead();
			while (true) {
				final IMessage msg = readMessage();
				System.out.println("[" + System.currentTimeMillis() + "] "
						+ "[READ  : " + peer.getName() + "] " + msg.toString());
				if (msg instanceof EndOfStream) {
					TCF.disconnect(peer);
					break;
				}
				taskUtilities.invokeLater(new Runnable() {
					public void run() {
						channelInterface.handleMessage(msg);
					}
				});
			}
			terminateRead();
		} catch (final Throwable x) {
			taskUtilities.invokeLater(new Runnable() {
				public void run() {
					taskUtilities.error(x);
				}
			});
		}
	}

	private void outProcess() {
		assert Thread.currentThread() == outThread;
		try {
			initWrite();
			IMessage msg;
			while (true) {
				msg = null;
				synchronized (outMessages) {
					while (outMessages.isEmpty())
						outMessages.wait();
					msg = outMessages.removeFirst();
				}
				System.out.println("[" + System.currentTimeMillis() + "] "
						+ "[WRITE : " + peer.getName() + "] " + msg.toString());
				writeMessage(msg);
				if (msg instanceof EndOfStream)
					break;
			}
			terminateWrite();
		} catch (final Throwable x) {
			taskUtilities.invokeLater(new Runnable() {
				public void run() {
					taskUtilities.error(x);
				}
			});
		}
	}

	public void sendMessage(IMessage message) {
		assert taskUtilities.isDispatchThread();
		synchronized (outMessages) {
			outMessages.addLast(message);
			outMessages.notify();
		}
	}

}