package org.eclipse.tm.tcf.internal;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.tm.tcf.extender.channel.IChannelInterface;
import org.eclipse.tm.tcf.extender.message.Command;
import org.eclipse.tm.tcf.extender.message.Event;
import org.eclipse.tm.tcf.extender.message.IMessage;
import org.eclipse.tm.tcf.extender.message.Progress;
import org.eclipse.tm.tcf.extender.message.Result;
import org.eclipse.tm.tcf.extender.peer.IPeer;
import org.eclipse.tm.tcf.extender.service.ITaskUtilities.ReturnRunnable;
import org.eclipse.tm.tcf.extender.service.provider.IServiceProviderInterface;
import org.eclipse.tm.tcf.extender.service.provider.IServiceProviderManager;
import org.eclipse.tm.tcf.extender.service.proxy.IServiceProxyInterface;
import org.eclipse.tm.tcf.extender.service.proxy.IServiceProxyManager;

public class PeerCommunication implements IServiceProviderInterface,
		IServiceProxyInterface, IChannelInterface {

	public static final int WAITING_TIME = 500;

	protected long tokenCnt;

	protected IPeer peer;

	protected Map<String, IEventListener> eventListeners = new HashMap<String, IEventListener>();
	protected Map<String, ICommandListener> commandListeners = new HashMap<String, ICommandListener>();
	protected Map<String, ICommandServer> commandServers = new HashMap<String, ICommandServer>();

	protected Collection<IMessageSender> messageSenders = new ArrayList<IMessageSender>();

	protected ValueAdderCommunication valueAdderCom;

	public PeerCommunication(IPeer peer) {

		assert TCFManager.getTaskManager().isDispatchThread();

		this.peer = peer;

		String[] types = new String[0];

		types = peer.getServiceValueAdderTypes();
		if (types != null && types.length != 0) {
			valueAdderCom = new ValueAdderCommunication(peer);
		} else {
			valueAdderCom = null;
		}

		types = new String[0];

		types = peer.getServiceProviderTypes();
		for (String type : types) {
			for (IServiceProviderManager manager : TCFManager
					.getServiceProviderManagers().get(type)) {
				manager.createServiceProvider(peer, this);
			}
		}

		types = new String[0];

		types = peer.getServiceProxyTypes();
		for (String type : types) {
			for (IServiceProxyManager manager : TCFManager
					.getServiceProxyManagers().get(type)) {
				manager.createServiceProxy(peer, this);
			}
		}

		if (valueAdderCom == null) {
			TCFManager.getChannelManagers().get(peer.getChannelType())
					.createChannel(peer, this);
		} else {
			valueAdderCom.createChannel(peer, this);
		}

	}

	protected PeerCommunication() {

	}

	/* UTILITIES */

	protected void error(Exception e) {
		assert TCFManager.getTaskManager().isDispatchThread();
		// TODO modify the error management
		// throw e;
	}

	protected void terminate() {
		assert TCFManager.getTaskManager().isDispatchThread();

		// TODO manage unfinished commands

		String[] types = new String[0];

		for (String token : commandListeners.keySet()) {
			ICommandListener listener = commandListeners.remove(token);
			listener.terminated(token, new Exception("Channel is closed"));
		}
		types = peer.getServiceProxyTypes();
		for (String type : types) {
			for (IServiceProxyManager manager : TCFManager
					.getServiceProxyManagers().get(type)) {
				manager.removeServiceProxy(peer);
			}
		}

		types = new String[0];

		types = peer.getServiceProviderTypes();
		for (String type : types) {
			for (IServiceProviderManager manager : TCFManager
					.getServiceProviderManagers().get(type)) {
				manager.removeServiceProvider(peer);
			}
		}

		if (valueAdderCom == null) {
			TCFManager.getChannelManagers().get(peer.getChannelType())
					.removeChannel(peer);
		} else {
			valueAdderCom.terminate();
		}

		valueAdderCom = null;

		peer = null;
	}

	protected String generateToken() {
		assert TCFManager.getTaskManager().isDispatchThread();
		return Long.toString(tokenCnt++);
	}

	/* ROUTING */

	protected void handleMessageFromTarget(final IMessage msg) {
		assert TCFManager.getTaskManager().isDispatchThread();
		switch (msg.getType()) {
		case Command.TYPE:
			ICommandServer cServer = commandServers.get(msg.getService());
			if (cServer == null)
				break;
			cServer.command(msg.getToken(), msg.getName(), msg.getData());
			return;
		case Event.TYPE:
			IEventListener eListener = eventListeners.get(msg.getService());
			if (eListener == null)
				break;
			eListener.event(msg.getName(), msg.getData());
			return;
		case Result.TYPE:
			ICommandListener cRListener = commandListeners.remove(msg
					.getToken());
			if (cRListener == null)
				break;
			cRListener.result(msg.getToken(), msg.getData());
			return;
		case Progress.TYPE:
			ICommandListener cPListener = commandListeners.get(msg.getToken());
			if (cPListener == null)
				break;
			cPListener.progress(msg.getToken(), msg.getData());
			return;
		default:
			error(null);
			break;
		}

	}

	protected void handleMessageToTarget(final IMessage msg) {
		assert TCFManager.getTaskManager().isDispatchThread();
		for (IMessageSender sender : messageSenders) {
			sender.sendMessage(msg);
		}
	}

	/* PROXY */

	public String sendCommand(final String serviceId, final String commandName,
			final Object[] data, final ICommandListener commandListener) {
		if (!TCFManager.getTaskManager().isDispatchThread()) {
			Object value = TCFManager.getTaskManager().invokeAndWaitReturn(
					WAITING_TIME, new ReturnRunnable() {
						public Object runAndReturn() throws Exception {
							return sendCommand(serviceId, commandName, data,
									commandListener);
						}
					});
			return (String) value;
		} else {
			String token = generateToken();
			if (commandListener != null)
				commandListeners.put(token, commandListener);
			IMessage msg = new Command(token, serviceId, commandName, data);
			handleMessageToTarget(msg);
			return token;
		}
	}

	public void addEventListener(final String serviceId,
			final IEventListener eventListener) {
		if (!TCFManager.getTaskManager().isDispatchThread()) {
			TCFManager.getTaskManager().invokeLater(new Runnable() {
				public void run() {
					addEventListener(serviceId, eventListener);
				}
			});
		} else {
			eventListeners.put(serviceId, eventListener);
		}
	}

	public void removeEventListener(final String serviceId,
			final IEventListener eventListener) {
		if (!TCFManager.getTaskManager().isDispatchThread()) {
			TCFManager.getTaskManager().invokeLater(new Runnable() {
				public void run() {
					removeEventListener(serviceId, eventListener);
				}
			});
		} else {
			eventListeners.remove(serviceId);
		}
	}

	/* PROVIDER */

	public void addCommandServer(final String serviceId,
			final ICommandServer server) {
		if (!TCFManager.getTaskManager().isDispatchThread()) {
			TCFManager.getTaskManager().invokeLater(new Runnable() {
				public void run() {
					addCommandServer(serviceId, server);
				}
			});
		} else {
			commandServers.put(serviceId, server);
		}
	}

	public void removeCommandServer(final String serviceId,
			final ICommandServer server) {
		if (!TCFManager.getTaskManager().isDispatchThread()) {
			TCFManager.getTaskManager().invokeLater(new Runnable() {
				public void run() {
					removeCommandServer(serviceId, server);
				}
			});
		} else {
			commandServers.remove(serviceId);
		}
	}

	public void sendEvent(final String serviceId, final String name,
			final Object[] data) {
		if (!TCFManager.getTaskManager().isDispatchThread()) {
			TCFManager.getTaskManager().invokeLater(new Runnable() {
				public void run() {
					sendEvent(serviceId, name, data);
				}
			});
		} else {
			String token = generateToken();
			IMessage msg = new Event(token, serviceId, name, data);
			handleMessageToTarget(msg);
		}
	}

	public void sendProgress(final String token, final Object[] data) {
		if (!TCFManager.getTaskManager().isDispatchThread()) {
			TCFManager.getTaskManager().invokeLater(new Runnable() {
				public void run() {
					sendProgress(token, data);
				}
			});
		} else {
			IMessage msg = new Progress(token, null, null, data);
			handleMessageToTarget(msg);
		}
	}

	public void sendResult(final String token, final Object[] data) {
		if (!TCFManager.getTaskManager().isDispatchThread()) {
			TCFManager.getTaskManager().invokeLater(new Runnable() {
				public void run() {
					sendResult(token, data);
				}
			});
		} else {
			IMessage msg = new Result(token, null, null, data);
			handleMessageToTarget(msg);
		}
	}

	/* CHANNEL */

	public void addMessageSender(final IMessageSender sender) {
		if (!TCFManager.getTaskManager().isDispatchThread()) {
			TCFManager.getTaskManager().invokeLater(new Runnable() {
				public void run() {
					addMessageSender(sender);
				}
			});
		} else {
			messageSenders.add(sender);
		}

	}

	public void handleMessage(final IMessage message) {
		if (!TCFManager.getTaskManager().isDispatchThread()) {
			TCFManager.getTaskManager().invokeLater(new Runnable() {
				public void run() {
					handleMessage(message);
				}
			});
		} else {
			handleMessageFromTarget(message);
		}

	}

	public void removeMessageSender(final IMessageSender sender) {
		if (!TCFManager.getTaskManager().isDispatchThread()) {
			TCFManager.getTaskManager().invokeLater(new Runnable() {
				public void run() {
					removeMessageSender(sender);
				}
			});
		} else {
			messageSenders.remove(sender);
		}
	}

}
